/* ------------------------------------------------------------------------- */
/*
 *  check_interval.cpp
 *
 *  Copyright (c) 2004 - 2008, clown. All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Last-modified: Tue 29 Jul 2008 13:21:03 JST
 */
/* ------------------------------------------------------------------------- */
#if defined(_WIN32) || defined(__WIN32__) || defined(WIN32)
#define CLX_USE_HIGH_PRECISION_TIMER
#endif

#include <memory>
#include <iostream>
#include <vector>
#include <cstring>
#include "clx/format.h"
#include "clx/argument.h"
#include "clx/stats.h"
#include "clx/version.h"
#include "icmp_msocket.h"

#define MAX_BURST 2
#define COOLING_TIME 5.0

/* ------------------------------------------------------------------------- */
//  user parameters
/* ------------------------------------------------------------------------- */
double interval_ = 100.0; // msec
double timeout_ = 10.0; // sec
int payload_ = 1450;
int probe_ = 10;
bool verbose_ = false;

/* ------------------------------------------------------------------------- */
//  usage
/* ------------------------------------------------------------------------- */
template <class Ch, class Tr>
void usage(std::basic_ostream<Ch, Tr>& sout) {
	sout << "usage: check_interval hostname [options]" << std::endl;
	sout << std::endl;
	sout << "options:" << std::endl;
	sout << "  -i # (msec)\t sending interval" << std::endl;
	sout << "  -l # (byte)\t data length" << std::endl;
	sout << "  -n #\t\t number of probing packets" << std::endl;
	sout << "  -t # (sec)\t timeout value" << std::endl;
	sout << "  -v\t\t print debug information" << std::endl;
	sout << std::endl;
	sout << "Use CLX version " << CLX_VERSION_STR << ". ";
	sout << "See also <" << CLX_URI << '>' << std::endl;
	sout << "Report bugs to <" << CLX_MAILTO << '>' << std::endl;
}

/* ------------------------------------------------------------------------- */
//  main
/* ------------------------------------------------------------------------- */
int main(int argc, char* argv[]) {
	clx::argument arg(argc, argv);
	if (arg.head().empty()) {
		usage(std::cerr);
		std::exit(-1);
	}
	else if (arg.exist("h,help")) {
		usage(std::cout);
		return 0;
	}
	
	// parsing user options
	arg("i,interval", interval_)("l,length", payload_)
		("n", probe_)("t,timeout", timeout_);
	interval_ /= 1000.0;
	payload_ += sizeof(struct icmphdr);
	if (arg.exist("v,verbose")) verbose_ = true;
	
	try {
		clx::icmp::measure_socket* sock = new clx::icmp::measure_socket(arg.head().at(0));
		clx::icmp::measure_handler h(payload_ + 1024);
		sock->state(clx::icmp::measure_socket::state_send);
		
		clx::icmp::sockmanager sm;
		sm.add(sock, h);
		
		std::vector<double> sends;
		std::vector<double> recvs;
		std::auto_ptr<char> buf(new char[payload_]);
		while (probe_ > 0) {
			int n = (probe_ < MAX_BURST) ? probe_ : MAX_BURST;
			for (int i = 0; i < n; i++) {
				clx::icmp::create_packet(i, buf.get(), payload_);
				if (sock->send(buf.get(), payload_) < 0) break;
				sock->add(i);
				
				if (i + 1 == n) {
					sock->state(clx::icmp::measure_socket::state_measure);
					sm.start(timeout_);
					if (sock->received() < sock->timestamps().size()) {
						sock->state(clx::icmp::measure_socket::state_abort);
					}
				}
				else sm.start(interval_);
				
				if (sock->state() == clx::icmp::measure_socket::state_abort) {
					if (verbose_) std::cerr << "measurement trial was aborted" << std::endl;
					sock->reset();
					break;
				}
			}
			
			for (u_int16_t j = 1; j < static_cast<u_int16_t>(sock->timestamps().size()); j++) {
				double x = (sock->timestamps()[j].send - sock->timestamps()[j-1].send) * 1000.0;
				double r = (sock->timestamps()[j].recv - sock->timestamps()[j-1].recv) * 1000.0;
				
				if (x < 0.0 || r < 0.0) { // discard results
					if (verbose_) std::cerr << "negative value of send/recv interval" << std::endl;
					break;
				}
				sends.push_back(x);
				recvs.push_back(r);
			}
			
			// cooling time
			sock->reset();
			probe_ -= MAX_BURST;
			if (probe_ <= 0) break;
			if (verbose_) {
				std::cout << "sleeping " << COOLING_TIME << " sec... rest "
					<< probe_ << " packets" << std::endl;
			}
			clx::sleep(COOLING_TIME);
		}
		
		// print statistics
		if (recvs.size() > 2) {
			std::pair<double, double> ci = clx::confidence_interval(recvs.begin(), recvs.end(), 1.96);
			std::cout << clx::format("%d bytes from %s: sample %d send %.1lf ms recv %.1lf [ %.1lf : %.1lf ] ms")
				% payload_ % sock->from().ipaddr() % recvs.size()
				% clx::mean(sends.begin(), sends.end())
				% clx::mean(recvs.begin(), recvs.end())
				% ci.first % ci.second
			<< std::endl;
		}
	}
	catch (clx::socket_error& e) {
		std::cerr << e.what() << std::endl;
		std::exit(-1);
	}
	catch (clx::sockaddress_error& e) {
		std::cerr << e.what() << std::endl;
		std::exit(-1);
	}
	
	return 0;
}
