/* ------------------------------------------------------------------------- */
/*
 *  icmp_msocket.h
 *
 *  Copyright (c) 2004 - 2008, clown. All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Last-modified: Tue 29 Jul 2008 13:26:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_ICMP_MSOCKET_H
#define CLX_ICMP_MSOCKET_H

#include <map>
#include "clx/icmp.h"
#include "clx/sockmanager.h"

namespace clx {
	namespace icmp {
		/* ----------------------------------------------------------------- */
		//  timestamp
		/* ----------------------------------------------------------------- */
		struct timestamp {
			double send;
			double recv;
		};
		
		/* ----------------------------------------------------------------- */
		//  measure_socket
		/* ----------------------------------------------------------------- */
		class measure_socket : public clx::icmp::socket {
		public:
			typedef u_int16_t seq_type;
			typedef struct timestamp timestamp_type;
			typedef std::map<seq_type, timestamp_type> time_map;
			typedef time_map::size_type size_type;
			
			enum { state_abort = 0, state_idle, state_send, state_measure };
			
			measure_socket() : super(), state_(state_idle), recv_(0), v_(), t_() {}
			measure_socket(const std::string& host) :
				super(host), state_(state_idle), recv_(0), v_(), t_() {}
			virtual ~measure_socket() throw() {}
			
			bool add(seq_type seq) {
				timestamp_type tmp;
				tmp.send = t_.total_elapsed();
				
				std::pair<time_map::iterator, bool> ret;
				ret = v_.insert(time_map::value_type(seq, tmp));
				return ret.second;
			}
			
			bool update(seq_type seq) {
				double tmp = t_.total_elapsed();
				time_map::iterator it = v_.find(seq);
				if (it == v_.end()) { return false; }
				it->second.recv = tmp;
				recv_++;
				return true;
			}
			
			void reset() {
				v_.clear();
				recv_ = 0;
				state_ = state_idle;
			}
			
			void state(int x) { state_ = x; }
			
			int state() const { return state_; }
			size_type received() const { return recv_; }
			
			time_map& timestamps() { return v_; }
			
		private:
			typedef clx::icmp::socket super;
			
			int state_;
			size_type recv_;
			time_map v_;
			clx::timer t_;
		};
		
		/* ----------------------------------------------------------------- */
		//  measure_handler
		/* ----------------------------------------------------------------- */
		class measure_handler {
		public:
			typedef measure_socket socket_type;
			typedef char char_type;
			
			explicit measure_handler(int size = 2048) :
				buf_(new char_type[size]), size_(size) {}
			
			template <class T, class SockManager>
			bool operator()(T* s, SockManager& sm) {
				socket_type* sock = dynamic_cast<socket_type*>(s);
				
				std::memset(buf_.get(), 0, size_);
				int len = sock->recv(buf_.get(), size_);
				if (len <= 0) {
					sock->state(socket_type::state_abort);
					return false;
				}
				
				clx::icmp::packet_header hdr(buf_.get());
				socket_type::seq_type seq = hdr.icmp()->sequence;
				if (!sock->update(seq)) sock->state(socket_type::state_abort);
				if (sock->state() == socket_type::state_measure &&
					sock->timestamps().size() == sock->received()) {
					sm.stop();
					sock->state(socket_type::state_idle);
				}
				
				return true;
			}
			
		private:
			std::auto_ptr<char_type> buf_;
			int size_;
		};
		
		/* ----------------------------------------------------------------- */
		//  create_packet
		/* ----------------------------------------------------------------- */
		void create_packet(u_int16_t seq, char* dest, int n) {
			clx::icmp::packet_header hdr;
			hdr.icmp()->type = ICMP_ECHO_REQUEST;
			hdr.icmp()->sequence = seq;
			
			std::memset(dest, 'a', n);
			std::memcpy(dest, (char*)hdr.icmp(), hdr.icmp_size());
		}
	};
};

#endif // CLX_ICMP_MSOCKET_H
